#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File: options.py
#
# Copyright 2021 Costas Tyfoxylos
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
#

"""
options package.

Import all parts from options here

.. _Google Python Style Guide:
   http://google.github.io/styleguide/pyguide.html
"""

import functools

import click
from click_option_group import MutuallyExclusiveOptionGroup

from .validators import (validate_email,
                         validate_region,
                         validate_token,
                         validate_account_name,
                         validate_account_password,
                         validate_mfa_device_name,
                         validate_mfa_device_serial)

__author__ = '''Costas Tyfoxylos <ctyfoxylos@schubergphilis.com>'''
__docformat__ = '''google'''
__date__ = '''24-12-2021'''
__copyright__ = '''Copyright 2021, Costas Tyfoxylos'''
__license__ = '''MIT'''
__maintainer__ = '''Costas Tyfoxylos'''
__email__ = '''<ctyfoxylos@schubergphilis.com>'''
__status__ = '''Development'''  # "Prototype", "Development", "Production".


email_option = click.option('-e',
                            '--email',
                            'account_email',
                            required=True,
                            type=str,
                            callback=validate_email,
                            help='The email to use.')

device_name_option = click.option('-d',
                                  '--device-name',
                                  'device_name',
                                  required=True,
                                  default='root-account-mfa-device',
                                  type=str,
                                  callback=validate_mfa_device_name,
                                  help='The name of the virtual device. Defaults to "root-account-mfa-device"')

device_serial_option = click.option('-d',
                                    '--device-serial',
                                    'device_serial',
                                    required=True,
                                    type=str,
                                    callback=validate_mfa_device_serial,
                                    help='The serial of the virtual device in the form'
                                         ' of arn:aws:iam::ACCOUNTID:mfa/DEVICE_NAME.')

password_option = click.option('-p',
                               '--password',
                               'password',
                               hide_input=True,
                               prompt=True,
                               confirmation_prompt=True,
                               required=True,
                               type=str,
                               callback=validate_account_password,
                               help='The root password of the account.')

region_option = click.option('-r',
                             '--region',
                             'region',
                             required=True,
                             type=str,
                             envvar='AWS_DEFAULT_REGION',
                             callback=validate_region,
                             help=('The home region of the account, can read from "AWS_DEFAULT_REGION" environment '
                                   'variable'))

captcha_option = click.option('-t',
                              '--2captcha-token',
                              'captcha_token',
                              required=False,
                              envvar='TWO_CAPTCHA_API_TOKEN',
                              callback=validate_token,
                              help='A valid token for the 2captcha service if automatic captcha solving is required. '
                                   'Can read from "TWO_CAPTCHA_API_TOKEN" environment variable')

account_name_option = click.option('-n',
                                   '--name',
                                   'account_name',
                                   required=True,
                                   type=str,
                                   callback=validate_account_name,
                                   help='The name of the account.')


def common_options(function):
    """Options common to all commands."""
    logging_options = MutuallyExclusiveOptionGroup('Logging options',
                                                   help='Sets the level of logging interactively or accepts a '
                                                        'configuration file.')
    options = [logging_options.option('-L',
                                      '--log-config',
                                      'log_config',
                                      type=click.File(),
                                      help='A config file for logging, mutually exclusive with setting the logging '
                                           'level interactively.'),
               logging_options.option('-l',
                                      '--log-level',
                                      'log_level',
                                      type=click.Choice(['debug', 'info', 'warning', 'error']),
                                      default='info',
                                      help='Provide the log level. Defaults to info. Mutually exclusive with providing '
                                           'a logging configuration file.')]
    return functools.reduce(lambda x, option: option(x), options, function)


def common_account_manager_options(function):
    """Options common to account manager commands."""
    options = [email_option,
               password_option,
               region_option,
               click.option('-m',
                            '--mfa-seed',
                            'mfa_serial',
                            required=False,
                            type=str,
                            help='The original mfa seed of the account if virtual MFA is enabled.'),
               captcha_option]
    return functools.reduce(lambda x, option: option(x), options, function)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File: actions.py
#
# Copyright 2021 Costas Tyfoxylos
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
#

"""
actions package.

Import all parts from actions here

.. _Google Python Style Guide:
   http://google.github.io/styleguide/pyguide.html
"""

from art import text2art
from awsapilib.captcha import Terminal
from awsapilib.console import (AccountManager,
                               PasswordManager)
from awsapilib.console.consoleexceptions import (UnableToResolveAccount,
                                                 UnableToRequestResetPassword)
from opnieuw import retry
from rich.console import Console
from rich.text import Text

__author__ = '''Costas Tyfoxylos <ctyfoxylos@schubergphilis.com>'''
__docformat__ = '''google'''
__date__ = '''24-12-2021'''
__copyright__ = '''Copyright 2021, Costas Tyfoxylos'''
__license__ = '''MIT'''
__maintainer__ = '''Costas Tyfoxylos'''
__email__ = '''<ctyfoxylos@schubergphilis.com>'''
__status__ = '''Development'''  # "Prototype", "Development", "Production".


def show_header():
    """Shows the project header."""
    console = Console()
    text = Text("AWS Account Lifecycle Manager")
    text.stylize("bold magenta")
    console.print(text2art('AWS ALF CLI'))
    console.print(text)


def get_account_manager_arguments(options):
    """Prepares the standard arguments for account manager actions."""
    args = {'email': options.get("account_email"),
            'password': options.get('password'),
            'region': options.get('region'),
            'mfa_serial': options.get('mfa_serial')}
    solver = options.get('solver')
    args['solver'] = solver if solver else Terminal()
    return args


def get_password_manager_arguments(options):
    """Prepares the standard arguments for password manager actions."""
    args = {}
    solver = options.get('solver')
    args['solver'] = solver if solver else Terminal()
    return args


@retry(retry_on_exceptions=UnableToResolveAccount, max_calls_total=5, retry_window_after_first_call_in_seconds=2)
def update_account_name(options, console):
    """Update an account's name.

    Args:
        options: The options provided by click
        console: The console provided by rich

    Returns:
        result (bool): True os success False on failure

    """
    args = get_account_manager_arguments(options)
    account_email = options.get("account_email")
    account_name = options.get('account_name')
    console.log('Please wait while account manager authenticates, it will retry on failure.')
    account_manager = AccountManager(**args)
    console.log(f"Updating account {account_email}'s name to {account_name}")
    result = account_manager.update_account_name(account_name)
    report_message = 'Succeeded' if result else 'Failed'
    console.log(f'{report_message} in updating name to {account_name}.')
    return result


@retry(retry_on_exceptions=UnableToResolveAccount, max_calls_total=5, retry_window_after_first_call_in_seconds=2)
def update_account_email(options, console):
    """Update an account's email.

    Args:
        options: The options provided by click
        console: The console provided by rich

    Returns:
        result (bool): True os success False on failure

    """
    args = get_account_manager_arguments(options)
    account_email = options.get("account_email")
    new_email = options.get('new_account_email')
    console.log('Please wait while account manager authenticates, it will retry on failure.')
    account_manager = AccountManager(**args)
    console.log(f'Updating account {account_email} to {new_email}')
    result = account_manager.update_account_email(new_email)
    report_message = 'Succeeded' if result else 'Failed'
    console.log(f'{report_message} in updating email to {new_email}.')
    return result


@retry(retry_on_exceptions=UnableToResolveAccount, max_calls_total=5, retry_window_after_first_call_in_seconds=2)
def terminate_account(options, console):
    """Terminates an account.

    Args:
        options: The options provided by click
        console: The console provided by rich

    Returns:
        result (bool): True os success False on failure

    """
    args = get_account_manager_arguments(options)
    account_email = options.get("account_email")
    console.log('Please wait while account manager authenticates, it will retry on failure.')
    account_manager = AccountManager(**args)
    console.log(f'Terminating account {account_email}')
    result = account_manager.terminate_account()
    report_message = 'Succeeded' if result else 'Failed'
    console.log(f'{report_message} in terminating account {account_email}.')
    return result


@retry(retry_on_exceptions=(UnableToResolveAccount, UnableToRequestResetPassword),
       max_calls_total=5,
       retry_window_after_first_call_in_seconds=2)
def password_reset_request(options, console):
    """Requests a password reset link.

    Args:
        options: The options provided by click
        console: The console provided by rich

    Returns:
        result (bool): True os success False on failure

    """
    args = get_password_manager_arguments(options)
    if args.get('solver'):
        console.log('Please wait while password manager resolves the captchas, '
                    'it will retry on failure.')
    password_manager = PasswordManager(**args)
    console.log('Requesting password reset.')
    result = password_manager.request_password_reset(options.get("account_email"))
    report_message = 'Succeeded' if result else 'Failed'
    console.log(f'{report_message} in requesting password reset.')
    return result


@retry(retry_on_exceptions=UnableToResolveAccount, max_calls_total=5, retry_window_after_first_call_in_seconds=2)
def password_reset(options, console):
    """Resets an account root password.

    Args:
        options: The options provided by click
        console: The console provided by rich

    Returns:
        result (bool): True os success False on failure

    """
    args = get_password_manager_arguments(options)
    if args.get('solver'):
        console.log('Please wait while password manager resolves the captchas, '
                    'it will retry on failure.')
    password_manager = PasswordManager(**args)
    console.log('Resetting password.')
    result = password_manager.reset_password(options.get('reset_url'),
                                             options.get('password'))
    report_message = 'Succeeded' if result else 'Failed'
    console.log(f'{report_message} in resetting password.')
    return result


@retry(retry_on_exceptions=UnableToResolveAccount, max_calls_total=5, retry_window_after_first_call_in_seconds=2)
def activate_mfa(options, console):
    """Activates virtual MFA on an account.

    Args:
        options: The options provided by click
        console: The console provided by rich

    Returns:
        device (VirtualMFADevice): The device representation returned by the activation process.

    """
    device_name = options.get('device_name')
    args = get_account_manager_arguments(options)
    account_email = options.get("account_email")
    console.log('Please wait while account manager authenticates, it will retry on failure.')
    account_manager = AccountManager(**args)
    console.log(f'Activating MFA device {device_name} on account {account_email}.')
    account_manager.iam.billing_console_access = True
    device = account_manager.mfa.create_virtual_device(device_name)
    console.log(f'Succeeded in activating MFA device "{device_name}".')
    return device


@retry(retry_on_exceptions=UnableToResolveAccount, max_calls_total=5, retry_window_after_first_call_in_seconds=2)
def deactivate_mfa(options, console):
    """Deactivates the virtual MFA of an account.

    Args:
        options: The options provided by click
        console: The console provided by rich

    Returns:
        result (bool): True os success False on failure

    """
    device_serial = options.get('device_serial')
    args = get_account_manager_arguments(options)
    account_email = options.get("account_email")
    console.log('Please wait while account manager authenticates, it will retry on failure.')
    account_manager = AccountManager(**args)
    console.log(f'Deleting MFA device {device_serial} on account {account_email}.')
    mfa_deleted = account_manager.mfa.delete_virtual_device(device_serial)
    report_message = 'Succeeded' if mfa_deleted else 'Failed'
    console.log(f'{report_message} in deleting MFA device "{device_serial}".')
    return mfa_deleted


@retry(retry_on_exceptions=UnableToResolveAccount, max_calls_total=5, retry_window_after_first_call_in_seconds=2)
def activate_iam_billing(options, console):
    """Activates IAM access to the billing console.

    Args:
        options: The options provided by click
        console: The console provided by rich

    Returns:
        result (bool): True os success False on failure

    """
    args = get_account_manager_arguments(options)
    account_email = options.get("account_email")
    console.log('Please wait while account manager authenticates, it will retry on failure.')
    account_manager = AccountManager(**args)
    console.log(f'Activating IAM access on billing console of account {account_email}.')
    account_manager.iam.billing_console_access = True
    result = account_manager.iam.billing_console_access
    report_message = 'Succeeded' if result else 'Failed'
    console.log(f'{report_message} in activating IAM access.')
    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File: validation.py
#
# Copyright 2021 Costas Tyfoxylos
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
#

"""
validation package.

Import all parts from validation here

.. _Google Python Style Guide:
   http://google.github.io/styleguide/pyguide.html
"""

import re
import string
import random

import click
from awsapilib import ControlTower
from awsapilib.captcha import Captcha2, InvalidOrNoBalanceApiToken

__author__ = '''Costas Tyfoxylos <ctyfoxylos@schubergphilis.com>'''
__docformat__ = '''google'''
__date__ = '''24-12-2021'''
__copyright__ = '''Copyright 2021, Costas Tyfoxylos'''
__license__ = '''MIT'''
__maintainer__ = '''Costas Tyfoxylos'''
__email__ = '''<ctyfoxylos@schubergphilis.com>'''
__status__ = '''Development'''  # "Prototype", "Development", "Production".


ACCOUNT_NAME_MAXIMUM_LENGTH = 50
ACCOUNT_EMAIL_MINIMUM_LENGTH = 6
ACCOUNT_EMAIL_MAXIMUM_LENGTH = 64


def generate_random_string(length):
    """Generates a random string according to the length provided."""
    pool = string.ascii_letters + string.digits
    return ''.join(random.choice(pool) for _ in range(length))


def get_random_account_name(account_name):
    """Generates a random name up to the maximum size allowed."""
    length = ACCOUNT_NAME_MAXIMUM_LENGTH - len(account_name)
    if not length > 2:
        raise ValueError('Name is too long to generate a random name for it.')
    return account_name + generate_random_string(length)


def get_random_account_email(account_email):
    """Generates a random email up to the maximum size allowed."""
    length = ACCOUNT_EMAIL_MAXIMUM_LENGTH - len(account_email)
    if not length > 2:
        raise ValueError('Email is too long to generate a new random one for it.')
    host, domain = account_email.split('@')
    return f'{host}+{generate_random_string(length-1)}@{domain}'


def validate_email(ctx, param, value):
    """Validates an email option."""
    if not value:
        return value
    try:
        if not all([re.match(r"(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)", value),
                    ACCOUNT_EMAIL_MINIMUM_LENGTH <= len(value) <= ACCOUNT_EMAIL_MAXIMUM_LENGTH]):
            raise ValueError(value)
        return value
    except ValueError as msg:
        click.echo(f'Incorrect email given, it must be a valid email'
                   f' between 6 and 64 characters: {msg}')
        value = click.prompt(param.name)
        return validate_email(ctx, param, value)


def validate_region(ctx, param, value):
    """Validates a AWS region option."""
    try:
        if value not in ControlTower.get_available_regions():
            raise ValueError(value)
        return value
    except ValueError as msg:
        click.echo(f'Incorrect {param.name} given: {msg}')
        value = click.prompt(param.name)
        return validate_region(ctx, param, value)


def validate_arn(ctx, param, value):
    """Validates an AWS ARN option."""
    conditions = [value.startswith('arn:aws:iam::'),
                  value[13:25].isnumeric(),
                  value[25:31] == ':role/']
    try:
        if not all(conditions):
            raise ValueError(value)
        return value
    except ValueError as msg:
        click.echo(f'Incorrect {param.name} given: {msg}')
        value = click.prompt(param.name)
        return validate_arn(ctx, param, value)


def validate_token(ctx, param, value):  # pylint: disable=unused-argument
    """Validates a 2Captcha Token option and inserts an active 2Captcha instance in the options."""
    if not value:
        return None
    try:
        captcha = Captcha2(value)
        _ = captcha.solver.balance()
    except InvalidOrNoBalanceApiToken:
        raise click.BadParameter('The api token provided is not valid or no balance left of the account.')
    ctx.params['solver'] = captcha
    return value


def validate_reset_link(ctx, param, value):  # pylint: disable=unused-argument
    """Validates a reset link."""
    valid_prefix = 'https://signin.aws.amazon.com/resetpassword?type=RootUser&token='
    if not value.startswith(valid_prefix):
        raise click.BadParameter('The reset link provided is not valid.')
    return value


def validate_account_name(ctx, param, value):
    """Validates an account name."""
    try:
        if not all([re.match(r"[\s\S]*", value), 1 <= len(value) <= ACCOUNT_NAME_MAXIMUM_LENGTH]):
            raise ValueError(value)
        return value
    except ValueError:
        click.echo('Invalid account name given, only ASCII characters are allowed with a length between 1 and 50.')
        value = click.prompt(param.name)
        return validate_account_name(ctx, param, value)


def validate_account_password(ctx, param, value):
    """Validates an account password."""
    lower = set(string.ascii_lowercase)
    upper = set(string.ascii_uppercase)
    digits = set(string.digits)
    punctuation_string = '! @ # $ % ^ & * ( ) < > [ ] { } | _ + - ='
    punctuation = set(punctuation_string.split())
    all_characters = set(string.ascii_lowercase +
                         string.ascii_uppercase +
                         string.digits +
                         punctuation_string.replace(' ', ''))
    try:
        matches = [int(bool(group.intersection(value))) for group in (lower, upper, digits, punctuation)]
        conditions = [sum(matches) >= 3,
                      8 <= len(value) <= 128,
                      all_characters.issuperset(set(value))]
        if not all(conditions):
            raise ValueError
        return value
    except ValueError:
        print(f'Invalid password provided, password should be ascii lower, upper and digits and one of '
              f'{punctuation_string} and it should match at least three of these categories and be between 8 and 128 '
              f'characters.')
        value = click.prompt(param.name)
        return validate_account_password(ctx, param, value)


def validate_mfa_device_name(ctx, param, value):
    """Validates a virtual MFA device name."""
    try:
        if not all([re.match(r"[\w+=/:,.@-]+", value), 9 <= len(value) <= 256]):
            raise ValueError(value)
        return value
    except ValueError:
        click.echo('Invalid device name given, only ASCII characters and + = / : , . @ - are allowed with a length '
                   'between 9 and 256 with no spaces.')
        value = click.prompt(param.name)
        return validate_mfa_device_name(ctx, param, value)


def validate_mfa_device_serial(ctx, param, value):
    """Validates a reset link."""
    conditions = [value.startswith('arn:aws:iam::'),
                  value[13:25].isnumeric(),
                  value[25:30] == ':mfa/']
    try:
        if not all(conditions):
            raise ValueError(value)
        return value
    except ValueError as msg:
        click.echo(f'Incorrect {param.name} given: {msg}')
        value = click.prompt(param.name)
        return validate_mfa_device_serial(ctx, param, value)
